BASE = "https://api.themoviedb.org"
V = "3"


ACCOUNT_DETAILS = f"{BASE}/{V}/account"
ACCOUNT_CREATED_LISTS = f"{ACCOUNT_DETAILS}/{{account_id}}/lists"
ACCOUNT_FAVORITE_MOVIES = f"{ACCOUNT_DETAILS}/{{account_id}}/favorite/movies"
ACCOUNT_FAVORITE_SHOWS = f"{ACCOUNT_DETAILS}/{{account_id}}/favorite/tv"
ACCOUNT_MARK_AS_FAVORITE = f"{ACCOUNT_DETAILS}/{{account_id}}/favorite"
ACCOUNT_RATED_MOVIES = f"{ACCOUNT_DETAILS}/{{account_id}}/rated/movies"
ACCOUNT_RATED_SHOWS = f"{ACCOUNT_DETAILS}/{{account_id}}/rated/tv"
ACCOUNT_RATED_EPISODES = f"{ACCOUNT_DETAILS}/{{account_id}}/rated/tv/episodes"
ACCOUNT_MOVIE_WATCHLIST = f"{ACCOUNT_DETAILS}/{{account_id}}/watchlist/movies"
ACCOUNT_SHOW_WATCHLIST = f"{ACCOUNT_DETAILS}/{{account_id}}/watchlist/tv"
ACCOUNT_ADD_TO_WATCHLIST = f"{ACCOUNT_DETAILS}/{{account_id}}/watchlist"


AUTH_GUEST_SESSION = f"{BASE}/{V}/authentication/guest_session/new"
AUTH_NEW_TOKEN = f"{BASE}/{V}/authentication/token/new"
AUTH_NEW_SESSION = f"{BASE}/{V}/authentication/session/new"
AUTH_VALIDATE_WITH_LOGIN = (
    f"{BASE}/{V}/authentication/token/validate_with_login"
)
AUTH_DELETE_SESSION = f"{BASE}/{V}/authentication/session"


SEARCH_MOVIE = f"{BASE}/{V}/search/movie"
SEARCH_SHOW = f"{BASE}/{V}/search/tv"
SEARCH_PERSON = f"{BASE}/{V}/search/person"
SEARCH_COMPANY = f"{BASE}/{V}/search/company"


DISCOVER_MOVIES = f"{BASE}/{V}/discover/movie"
DISCOVER_SHOWS = f"{BASE}/{V}/discover/tv"


FIND = f"{BASE}/{V}/find/{{external_id}}"


MOVIE_CERTIFICATION = f"{BASE}/{V}/certification/movie/list"
SHOW_CERTIFICATION = f"{BASE}/{V}/certification/tv/list"
IMAGE_CONFIGURATION = f"{BASE}/{V}/configuration"
COUNTRIES_CONFIGURATION = f"{BASE}/{V}/configuration/countries"
JOBS_CONFIGURATION = f"{BASE}/{V}/configuration/jobs"
LANGUAGES_CONFIGURATION = f"{BASE}/{V}/configuration/languages"
PRIMARY_TRANSLATIONS_CONFIGURATION = (
    f"{BASE}/{V}/configuration/primary_translations"
)
TIMEZONES_CONFIGURATION = f"{BASE}/{V}/configuration/timezones"


MOVIE_DETAILS = f"{BASE}/{V}/movie/{{movie_id}}"
MOVIE_ALTERNATIVE_TITLES = f"{MOVIE_DETAILS}/alternative_titles"
MOVIE_CHANGES = f"{MOVIE_DETAILS}/changes"
MOVIE_CREDITS = f"{MOVIE_DETAILS}/credits"
MOVIE_EXTERNAL_IDS = f"{MOVIE_DETAILS}/external_ids"
MOVIE_IMAGES = f"{MOVIE_DETAILS}/images"
MOVIE_KEYWORDS = f"{MOVIE_DETAILS}/keywords"
MOVIE_RELEASE_DATES = f"{MOVIE_DETAILS}/release_dates"
MOVIE_VIDEOS = f"{MOVIE_DETAILS}/videos"
MOVIE_TRANSLATIONS = f"{MOVIE_DETAILS}/translations"
MOVIE_RECOMMENDATIONS = f"{MOVIE_DETAILS}/recommendations"
MOVIE_SIMILAR = f"{MOVIE_DETAILS}/similar"
MOVIE_REVIEWS = f"{MOVIE_DETAILS}/reviews"
MOVIE_LISTS = f"{MOVIE_DETAILS}/lists"
RATE_MOVIE = f"{MOVIE_DETAILS}/rating"
DELETE_MOVIE_RATING = f"{MOVIE_DETAILS}/rating"
ALL_MOVIE_SECOND_SUFFIXES = [
    "alternative_titles",
    "changes",
    "credits",
    "external_ids",
    "images",
    "keywords",
    "release_dates",
    "videos",
    "translations",
    "recommendations",
    "similar",
    "reviews",
    "lists",
]


SHOW_DETAILS = f"{BASE}/{V}/tv/{{show_id}}"
SHOW_ALTERNATIVE_TITLES = f"{SHOW_DETAILS}/alternative_titles"
SHOW_CHANGES = f"{SHOW_DETAILS}/changes"
SHOW_CONTENT_RATINGS = f"{SHOW_DETAILS}/content_ratings"
SHOW_CREDITS = f"{SHOW_DETAILS}/credits"
SHOW_EPISODE_GROUPS = f"{SHOW_DETAILS}/episode_groups"
SHOW_EXTERNAL_IDS = f"{SHOW_DETAILS}/external_ids"
SHOW_IMAGES = f"{SHOW_DETAILS}/images"
SHOW_KEYWORDS = f"{SHOW_DETAILS}/keywords"
SHOW_RECOMMENDATIONS = f"{SHOW_DETAILS}/recommendations"
SHOW_REVIEWS = f"{SHOW_DETAILS}/reviews"
SHOW_SCREENED_THEATRICALLY = f"{SHOW_DETAILS}/screened_theatrically"
SHOW_SIMILAR = f"{SHOW_DETAILS}/similar"
SHOW_TRANSLATIONS = f"{SHOW_DETAILS}/translations"
SHOW_VIDEOS = f"{SHOW_DETAILS}/videos"
RATE_SHOW = f"{SHOW_DETAILS}/rating"
DELETE_SHOW_RATING = f"{SHOW_DETAILS}/rating"
ALL_SHOW_SECOND_SUFFIXES = [
    "alternative_titles",
    "changes",
    "content_ratings",
    "credits",
    "episode_groups",
    "external_ids",
    "images",
    "keywords",
    "recommendations",
    "reviews",
    "screened_theatrically",
    "similar",
    "translations",
    "videos",
]


PERSON_DETAILS = f"{BASE}/{V}/person/{{person_id}}"
PERSON_CHANGES = f"{PERSON_DETAILS}/changes"
PERSON_MOVIE_CREDITS = f"{PERSON_DETAILS}/movie_credits"
PERSON_SHOW_CREDITS = f"{PERSON_DETAILS}/tv_credits"
PERSON_COMBINED_CREDITS = f"{PERSON_DETAILS}/combined_credits"
PERSON_EXTERNAL_IDS = f"{PERSON_DETAILS}/external_ids"
PERSON_IMAGES = f"{PERSON_DETAILS}/images"
PERSON_TAGGED_IMAGES = f"{PERSON_DETAILS}/tagged_images"
PERSON_TRANSLATIONS = f"{PERSON_DETAILS}/translations"
ALL_PERSON_SECOND_SUFFIXES = [
    "changes",
    "movie_credits",
    "tv_credits",
    "combined_credits",
    "external_ids",
    "images",
    "tagged_images",
    "translations",
]

SEASON_DETAILS = f"{BASE}/{V}/tv/{{show_id}}/season/{{season_number}}"
SEASON_CHANGES = f"{BASE}/{V}/tv/season/{{season_id}}/changes"
SEASON_CREDITS = f"{SEASON_DETAILS}/credits"
SEASON_EXTERNAL_IDS = f"{SEASON_DETAILS}/external_ids"
SEASON_IMAGES = f"{SEASON_DETAILS}/images"
SEASON_VIDEOS = f"{SEASON_DETAILS}/videos"
ALL_SEASON_SECOND_SUFFIXES = ["credits", "external_ids", "images", "videos"]


EPISODE_DETAILS = f"{BASE}/{V}/tv/{{show_id}}/season/{{season_number}}/episode/{{episode_number}}"
EPISODE_CHANGES = f"{BASE}/{V}/tv/{{episode_id}}/changes"
EPISODE_CREDITS = f"{EPISODE_DETAILS}/credits"
EPISODE_EXTERNAL_IDS = f"{EPISODE_DETAILS}/external_ids"
EPISODE_TRANSLATIONS = f"{EPISODE_DETAILS}/translations"
EPISODE_IMAGES = f"{EPISODE_DETAILS}/images"
EPISODE_VIDEOS = f"{EPISODE_DETAILS}/videos"
ALL_EPISODE_SECOND_SUFFIXES = [
    "credits",
    "external_ids",
    "translations",
    "images",
    "videos",
]

PERSON_DETAILS = f"{BASE}/{V}/person/{{person_id}}"
PERSON_CHANGES = f"{PERSON_DETAILS}/changes"
PERSON_MOVIE_CREDITS = f"{PERSON_DETAILS}/movie_credits"
PERSON_SHOW_CREDITS = f"{PERSON_DETAILS}/tv_credits"
PERSON_COMBINED_CREDITS = f"{PERSON_DETAILS}/combined_credits"
PERSON_EXTERNAL_IDS = f"{PERSON_DETAILS}/external_ids"
PERSON_IMAGES = f"{PERSON_DETAILS}/images"
PERSON_TAGGED_IMAGES = f"{PERSON_DETAILS}/tagged_images"
PERSON_TRANSLATIONS = f"{PERSON_DETAILS}/translations"
ALL_PERSON_SECOND_SUFFIXES = [
    "changes",
    "movie_credits",
    "tv_credits",
    "combined_credits",
    "external_ids",
    "images",
    "tagged_images",
    "translations",
]


COMPANY_DETAILS = f"{BASE}/{V}/company/{{company_id}}"
COMPANY_ALTERNATIVE_NAMES = f"{COMPANY_DETAILS}/alternative_names"
COMPANY_IMAGES = f"{COMPANY_DETAILS}/images"


KEYWORD_DETAILS = f"{BASE}/{V}/keyword/{{keyword_id}}"
KEYWORD_MOVIES = f"{KEYWORD_DETAILS}/movies"


MOVIE_GENRES = f"{BASE}/{V}/genre/movie/list"
SHOW_GENRES = f"{BASE}/{V}/genre/tv/list"


CREDIT_DETAILS = f"{BASE}/{V}/credit/{{credit_id}}"


LIST_DETAILS = f"{BASE}/{V}/list/{{list_id}}"
LIST_CHECK_MOVIE_STATUS = f"{LIST_DETAILS}/item_status"
CREATE_LIST = f"{BASE}/{V}/list"
DELETE_LIST = f"{BASE}/{V}/list/{{list_id}}"
ADD_MOVIE_TO_LIST = f"{BASE}/{V}/list/{{list_id}}/add_item"
REMOVE_MOVIE_FROM_LIST = f"{BASE}/{V}/list/{{list_id}}/remove_item"
CLEAR_LIST = f"{BASE}/{V}/list/{{list_id}}/clear"



from typing import Iterator, List, Optional, Tuple
from importlib import import_module

import isle._urls as URL
from ._tmdb import TMDb

from .._config import tmdb_api_key


__all__ = ["Show", "Season", "Episode"]


def _import_person():
    global Person
    Person = import_module("isle.objects._person").Person


def _import_image():
    global Image
    Image = import_module("isle.objects._others").Image


def _import_language():
    global Language
    Language = import_module("isle.objects._others").Language


def _import_country():
    global Country
    Country = import_module("isle.objects._others").Country


def _import_company():
    global Company
    Company = import_module("isle.objects._company").Company


def _import_vote():
    global Vote
    Vote = import_module("isle.objects._others").Vote


def _import_video():
    global Video
    Video = import_module("isle.objects._others").Video


def _import_genre():
    global Genre
    Genre = import_module("isle.objects._others").Genre


def _import_keyword():
    global Keyword
    Keyword = import_module("isle.objects._others").Keyword


def _import_credit():
    global Credit
    Credit = import_module("isle.objects._others").Credit


def _import_all():
    _import_person()
    _import_image()
    _import_language()
    _import_country()
    _import_company()
    _import_vote()
    _import_video()
    _import_genre()
    _import_keyword()
    _import_credit()


class Show(TMDb):
    """Represents a TV show."""

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        _import_all()

    def _init(self):
        return self.get_all()

    @property
    def title(self):
        """Return titles of a TV show in different languages.
        Each key is an ISO-3166-1 code (such as `"US"`, `"RU"`,
        etc.), and the value is the corresponding title.

        There is two special keys: `"original"` and `"default"`.
        The first one returns the original movie title and the
        second one depends on the location."""
        titles = {
            "default": self._getdata("name"),
            "original": self._getdata("original_name"),
        }
        for item in self._getdata("translations")["translations"]:
            code, title = item["iso_3166_1"], item["data"]["name"]
            if title:
                titles[code] = title
        return titles

    @property
    def overview(self):
        """Return the overviews of a TV show in different
        languages. Each key is an ISO-3166-1 code (such as `"US"`,
        `"RU"`, etc.), and the value is the corresponding overview.

        There is a key `"default"`. It depends on the location."""
        overviews = {"default": self._getdata("overview")}
        for item in self._getdata("translations")["translations"]:
            code, overview = item["iso_3166_1"], item["data"]["overview"]
            if overview:
                overviews[code] = overview
        return overviews

    @property
    def homepage(self):
        """Return a TV show homepages. Each key is an ISO-3166-1
        code (such as `"US"`, `"RU"`, etc.) and the value is the
        corresponding homepage.

        There is a key `"default"`. It depends on the location.
        """
        default = self._getdata("homepage")
        pages = {"default": default} if default else {}
        for item in self._getdata("translations")["translations"]:
            code, page = item["iso_3166_1"], item["data"]["homepage"]
            if page:
                pages[code] = page
        return pages

    @property
    def creators(self):
        """Return the creators of a TV show."""
        creators = []
        for item in self._getdata("created_by"):
            creators.append(Person(item["id"], **item))
        return creators

    @property
    def backdrops(self):
        """Return backdrops that belong to a TV show. Each item
        is an instance of the `Image` class."""
        backdrops = []
        for item in self._getdata("images")["backdrops"]:
            backdrops.append(Image(item, type_="backdrop"))
        return backdrops

    @property
    def posters(self):
        """Return posters that belong to a TV show. Each item is
        an instance of the `Image` class."""
        posters = []
        for item in self._getdata("images")["posters"]:
            posters.append(Image(item, type_="poster"))
        return posters

    @property
    def runtimes(self):
        """Return the running times of TV show episodes."""
        return self._getdata("episode_run_time")

    @property
    def first_air_date(self):
        """Return the air date of the first episode."""
        return self._getdata("first_air_date")

    @property
    def last_air_date(self):
        """Return the air date of the last episode."""
        return self._getdata("last_air_date")

    @property
    def in_production(self):
        """Return `True` if a TV show is in production."""
        return self._getdata("in_production")

    @property
    def languages(self):
        """Return the languages spoken in a TV show. Each item is
        an instance of the `Language` class."""
        iso_codes = self._getdata("languages")
        languages = []
        for code in iso_codes:
            try:
                item = self._all_languages[code]
            except AttributeError:
                self._all_languages = self._get_all_languages()
                item = self._all_languages[code]
            languages.append(
                Language(
                    iso_639_1=code,
                    english_name=item["english_name"],
                    original_name=item["name"],
                )
            )
        return languages

    @property
    def last_episode(self):
        """Return the last episode."""
        item = self._getdata("last_episode_to_air")
        return Episode(
            item["episode_number"],
            show_id=self.tmdb_id,
            season_number=item["season_number"],
        )

    @property
    def next_episode(self):
        """Return the next episode."""
        item = self._getdata("next_episode_to_air")
        if item:
            return Episode(
                item["episode_number"],
                show_id=self.tmdb_id,
                season_number=item["season_number"],
            )
        else:
            return item

    @property
    def n_episodes(self):
        """Return the number of episodes."""
        return self._getdata("number_of_episodes")

    @property
    def n_seasons(self):
        """Return the number of seasons."""
        return self._getdata("number_of_seasons")

    @property
    def countries(self):
        """Return production countries. Each item is an instance of
        the `Country` class."""
        countries = []
        for code in self._getdata("origin_country"):
            if code:
                try:
                    english_name = self._all_countries[code]
                except AttributeError:
                    self._all_countries = self._get_all_countries()
                    english_name = self._all_countries[code]
                countries.append(
                    Country(iso_3166_1=code, english_name=english_name)
                )
            else:
                continue
        return countries

    @property
    def popularity(self):
        """Return popularity of a TV show on TMDb.

        See more:
        https://developers.themoviedb.org/3/getting-started/popularity."""
        return self._getdata("popularity")

    @property
    def companies(self):
        """Return the production companies. Each item is an
        instance of the `Company` class."""
        companies = []
        for item in self._getdata("production_companies"):
            companies.append(Company(item["id"], **item))
        return companies

    @property
    def seasons(self):
        """Return seasons. Each item is an instance of the `Season`
        class."""
        seasons = []
        for item in self._getdata("seasons"):
            seasons.append(
                Season(item["season_number"], show_id=self.tmdb_id, **item)
            )
        return seasons

    @property
    def status(self):
        return self._getdata("status")

    @property
    def type(self):
        return self._getdata("type")

    @property
    def vote(self):
        """Return an instance of the `Vote` class. It is like a
        `NamedTuple` object with two attributes: `average` and
        `count`."""
        average = self._getdata("vote_average")
        count = self._getdata("vote_count")
        return Vote(average=average, count=count)

    @property
    def videos(self):
        """Return videos. Each item is an instance of the `Video`
        class."""
        videos = []
        for item in self._getdata("videos")["results"]:
            url = f"https://www.youtube.com/watch?v={item['key']}"
            videos.append(Video(name=item["name"], type=item["type"], url=url))
        return videos

    @property
    def genres(self):
        """Return a TV show genres."""
        genres = []
        for item in self._getdata("genres"):
            genres.append(Genre(tmdb_id=item["id"], name=item["name"]))
        return genres

    @property
    def keywords(self):
        """Return a TV show keywords."""
        keywords = []
        for item in self._getdata("keywords")["results"]:
            keywords.append(Keyword(item["id"], **item))
        return keywords

    @property
    def imdb_id(self):
        """Return the ID of a TV show on IMDb."""
        return self._getdata("external_ids")["imdb_id"]

    @property
    def tvdb_id(self):
        """Return the ID of a TV show on TVDb."""
        return self._getdata("external_ids")["tvdb_id"]

    @property
    def facebook_id(self):
        """Return the ID of a TV show on Facebook."""
        return self._getdata("external_ids")["facebook_id"]

    @property
    def instagram_id(self):
        """Return the ID of a TV show on Instagram."""
        return self._getdata("external_ids")["instagram_id"]

    @property
    def twitter_id(self):
        """Return the ID of a TV show on Twitter."""
        return self._getdata("external_ids")["twitter_id"]

    @property
    def ratings(self):
        """Return content ratings (certifications) that have been
        added to a TV show. Each key is an ISO-3166-1
        code (such as `"US"`, `"RU"`, etc.) and the value is the
        corresponding rating."""
        ratings = {}
        for item in self._getdata("content_ratings")["results"]:
            ratings[item["iso_3166_1"]] = item["rating"]
        return ratings

    @property
    def cast(self):
        """Return TV show cast as list of tuples
        `(Person, Credit)`."""
        cast = []
        for item in self._getdata("credits")["cast"]:
            kwargs = {
                "credit_type": "cast",
                "media_type": "tv",
                "department": "Acting",
                "job": "Actor",
            }
            credit = Credit(
                item["credit_id"],
                media_data=self.data,
                character=item["character"],
                **kwargs,
            )
            person = Person(item["id"], **item)
            cast.append((person, credit))
        return cast

    @property
    def crew(self):
        """Return TV show crew as list of tuples
        `(Person, Credit)`."""
        crew = []
        for item in self._getdata("credits")["crew"]:
            kwargs = {
                "credit_type": "crew",
                "media_type": "tv",
                "department": item["department"],
                "job": item["job"],
            }
            credit = Credit(item["credit_id"], media_data=self.data, **kwargs)
            person = Person(item["id"], **item)
            crew.append((person, credit))
        return crew

    def _get_all_countries(self):
        data = self._request(
            URL.COUNTRIES_CONFIGURATION, **{"api_key": tmdb_api_key()}
        )
        countries = {}
        for item in data:
            countries[item["iso_3166_1"]] = item["english_name"]
        return countries

    def get_all(self, **params):
        """Get all information about a TV show. This method
        makes only one API request."""
        methods = ",".join(URL.ALL_SHOW_SECOND_SUFFIXES)
        all_data = self.get_details(
            **{"append_to_response": methods, **params}
        )
        self.data.update(all_data)
        return all_data

    def get_details(self, **params):
        """Get the primary information about a TV show."""
        details = self._request(
            URL.SHOW_DETAILS.format(show_id=self.tmdb_id), **params
        )
        self.data.update(details)
        return details

    def get_alternative_titles(self, **params):
        """Get all of the alternative titles for a TV show."""
        alternative_titles = self._request(
            URL.SHOW_ALTERNATIVE_TITLES.format(show_id=self.tmdb_id), **params
        )
        self.data.update({"alternative_titles": alternative_titles})
        return alternative_titles

    def get_changes(self, **params):
        """Get the changes for a TV show. By default only the
        last 24 hours are returned."""
        changes = self._request(
            URL.SHOW_CHANGES.format(show_id=self.tmdb_id), **params
        )
        self.data.update({"changes": changes})
        return changes

    def get_content_ratings(self, **params):
        """Get the list of content ratings (certifications)
        that have been added to a TV show."""
        content_ratings = self._request(
            URL.SHOW_CONTENT_RATINGS.format(show_id=self.tmdb_id), **params
        )
        self.data.update({"content_ratings": content_ratings})
        return content_ratings

    def get_credits(self, **params):
        """Get the cast and crew for a TV show."""
        credits = self._request(
            URL.SHOW_CREDITS.format(show_id=self.tmdb_id), **params
        )
        self.data.update({"credits": credits})
        return credits

    def get_episode_groups(self, **params):
        """Get all of the episode groups that have been
        created for a TV show."""
        episode_groups = self._request(
            URL.SHOW_EPISODE_GROUPS.format(show_id=self.tmdb_id), **params
        )
        self.data.update({"episode_groups": episode_groups})
        return episode_groups

    def get_external_ids(self, **params):
        """Get the external ids for a TV show. Such as
        Facebook, Instagram, Twitter, IMDb and TVDB"""
        external_ids = self._request(
            URL.SHOW_EXTERNAL_IDS.format(show_id=self.tmdb_id), **params
        )
        self.data.update({"external_ids": external_ids})
        return external_ids

    def get_images(self, **params):
        """Get the images that belong to a TV show."""
        images = self._request(
            URL.SHOW_IMAGES.format(show_id=self.tmdb_id), **params
        )
        self.data.update({"images": images})
        return images

    def get_keywords(self, **params):
        """Get the keywords that have been added to a TV show."""
        keywords = self._request(
            URL.SHOW_KEYWORDS.format(show_id=self.tmdb_id), **params
        )
        self.data.update({"keywords": keywords})
        return keywords

    def iter_recommendations(self, **params):
        """Get a list of recommended movies for a TV show."""
        yield from self._iter_request(
            URL.SHOW_RECOMMENDATIONS.format(show_id=self.tmdb_id), **params
        )

    def iter_reviews(self, **params):
        """Get the user reviews for a TV show."""
        yield from self._iter_request(
            URL.SHOW_REVIEWS.format(show_id=self.tmdb_id), **params
        )

    def get_screened_theatrically(self, **params):
        """Get a list of seasons or episodes that have been
        screened in a film festival or theatre."""
        screened_theatrically = self._request(
            URL.SHOW_SCREENED_THEATRICALLY.format(show_id=self.tmdb_id),
            **params,
        )
        self.data.update({"screened_theatrically": screened_theatrically})
        return screened_theatrically

    def iter_similar_shows(self, **params):
        """Get a list of recommended movies for a TV shows."""
        yield from self._iter_request(
            URL.SHOW_SIMILAR.format(show_id=self.tmdb_id), **params
        )

    def get_translations(self, **params):
        """Get a list of the translations that exist for a
        TV show."""
        translations = self._request(
            URL.SHOW_TRANSLATIONS.format(show_id=self.tmdb_id), **params
        )
        self.data.update({"translations": translations})
        return translations

    def get_videos(self, **params):
        """Get the videos that have been added to a TV show."""
        videos = self._request(
            URL.SHOW_VIDEOS.format(show_id=self.tmdb_id), **params
        )
        self.data.update({"videos": videos})
        return videos


class Season(TMDb):
    """Represents a TV show season."""

    def __init__(self, n: int, *, show_id: int, **kwargs):
        _import_all()
        self.data = {"season_number": n, **kwargs}
        self.show_id = show_id
        self.number = self.data["season_number"]
        self.n_requests = 0

    def _init(self):
        self.get_all()

    @property
    def tmdb_id(self):
        """Return the ID of a season on TMDb."""
        return self._getdata("id")

    @property
    def n(self):
        """Return a season number."""
        return self._getdata("season_number")

    @property
    def title(self):
        """Return a season's name."""
        return self._getdata("name")

    @property
    def overview(self):
        """Return a season's overview."""
        return self._getdata("overview")

    @property
    def air_date(self):
        """Return the air date of a season."""
        return self._getdata("air_date")

    @property
    def episodes(self):
        """Return a season's episodes."""
        episodes = []
        for item in self._getdata("episodes"):
            episodes.append(
                Episode(
                    item["episode_number"],
                    show_id=self.show_id,
                    season_number=item["season_number"],
                )
            )
        return episodes

    @property
    def tvdb_id(self):
        """Return the ID of a season on TVDb."""
        return self._getdata("external_ids")["tvdb_id"]

    @property
    def posters(self):
        """Return poster images that belong to a season. Each item
        is an instance of the `Image` class."""

        def _i(i):
            return Image(i, type_="poster")

        return list(map(_i, self._getdata("images")["posters"]))

    @property
    def videos(self):
        """Return videos. Each item is an instance of the `Video`
        class."""
        videos = []
        for item in self._getdata("videos")["results"]:
            url = f"https://www.youtube.com/watch?v={item['key']}"
            videos.append(Video(name=item["name"], type=item["type"], url=url))
        return videos

    @property
    def cast(self):
        """Return TV season cast as list of tuples `(Person, Credit)`."""
        cast = []
        for item in self._getdata("credits")["cast"]:
            kwargs = {
                "credit_type": "cast",
                "media_type": "movie",
                "department": "Acting",
                "job": "Actor",
            }
            credit = Credit(
                item["credit_id"],
                media_data=self.data,
                character=item["character"],
                **kwargs,
            )
            person = Person(item["id"], **item)
            cast.append((person, credit))
        return cast

    @property
    def crew(self):
        """Return a TV season crew as list of tuples `(Person, Credit)`."""
        crew = []
        for item in self._getdata("credits")["crew"]:
            kwargs = {
                "credit_type": "crew",
                "media_type": "movie",
                "department": item["department"],
                "job": item["job"],
            }
            credit = Credit(item["credit_id"], media_data=self.data, **kwargs)
            person = Person(item["id"], **item)
            crew.append((person, credit))
        return crew

    def get_all(self, **params):
        """Get all information about a TV season in a single
        response."""
        methods = ",".join(URL.ALL_SEASON_SECOND_SUFFIXES)
        all_data = self.get_details(
            **{"append_to_response": methods, **params}
        )
        self.data.update(all_data)
        return all_data

    def get_details(self, **params):
        """Get the primary TV season details."""
        details = self._request(
            URL.SEASON_DETAILS.format(
                show_id=self.show_id, season_number=self.number
            ),
            **params,
        )
        self.data.update(details)
        return details

    def get_changes(self, **params):
        """Get the changes for a TV season. By default only the
        last 24 hours are returned."""
        changes = self._request(
            URL.SEASON_CHANGES.format(season_id=self.tmdb_id), **params
        )
        self.data.update({"changes": changes})
        return changes

    def get_credits(self, **params):
        """Get the credits for TV season."""
        movie_credits = self._request(
            URL.SEASON_CREDITS.format(
                show_id=self.show_id, season_number=self.number
            ),
            **params,
        )
        self.data.update({"credits": movie_credits})
        return movie_credits

    def get_external_ids(self, **params):
        """Get the external ids for a TV season."""
        external_ids = self._request(
            URL.SEASON_EXTERNAL_IDS.format(
                show_id=self.show_id, season_number=self.number
            ),
            **params,
        )
        self.data.update({"external_ids": external_ids})
        return external_ids

    def get_images(self, **params):
        """Get the images that belong to a TV season."""
        images = self._request(
            URL.SEASON_IMAGES.format(
                show_id=self.show_id, season_number=self.number
            ),
            **params,
        )
        self.data.update({"images": images})
        return images

    def get_videos(self, **params):
        """Get the videos that have been added to a TV season."""
        videos = self._request(
            URL.SEASON_VIDEOS.format(
                show_id=self.show_id, season_number=self.number
            ),
            **params,
        )
        self.data.update({"videos": videos})
        return videos


class Episode(TMDb):
    """Represents a TV show episode."""

    def __init__(self, n, *, show_id, season_number, **kwargs):
        _import_all()
        self.data = {
            "episode_number": n,
            "season_number": season_number,
            **kwargs,
        }
        self.show_id = show_id
        self.number = self.data["episode_number"]
        self.season_number = self.data["season_number"]
        self.n_requests = 0

    def _init(self):
        self.get_all()

    @property
    def tmdb_id(self):
        """Return the ID of an episode on TMDb."""
        return self._getdata("id")

    @property
    def tvdb_id(self):
        """Return the ID of an episode on TVDb."""
        return self._getdata("external_ids")["tvdb_id"]

    @property
    def imdb_id(self):
        """Return the ID of an episode on IMDb."""
        return self._getdata("external_ids")["imdb_id"]

    @property
    def air_date(self):
        """Return the air date of an episode."""
        return self._getdata("air_date")

    @property
    def n(self):
        """Return an episode number."""
        return self._getdata("episode_number")

    @property
    def sn(self):
        """Return a season number."""
        return self._getdata("season_number")

    @property
    def title(self):
        """Return titles of an episode in different languages.
        Each key is an ISO-3166-1 code (such as `"US"`, `"RU"`,
        etc.), and the value is the corresponding title.

        There is a key `"default"`. It depends on the location."""

        def _n(n):
            return n["iso_3166_1"], n["data"]["name"]

        names = {}
        names["default"] = self._getdata("name")
        for k, v in map(_n, self._getdata("translations")["translations"]):
            if v:
                names[k] = v
        return names

    @property
    def overview(self):
        """Return the overviews of an episode in different
        languages. Each key is an ISO-3166-1 code (such as `"US"`,
        `"RU"`, etc.), and the value is the corresponding overview.

        There is a key `"default"`. It depends on the location."""

        def _o(o):
            return o["iso_3166_1"], o["data"]["overview"]

        overviews = {}
        overviews["default"] = self._getdata("overview")
        for k, v in map(_o, self._getdata("translations")["translations"]):
            if v:
                overviews[k] = v
        return overviews

    @property
    def stills(self):
        """Return images that belong to an episode. Each item is
        an instance of the `Image` class."""

        def _i(i):
            return Image(i, type_="still")

        return list(map(_i, self._getdata("images")["stills"]))

    @property
    def videos(self):
        """Return videos. Each item is an instance of the `Video`
        class."""
        videos = []
        for item in self._getdata("videos")["results"]:
            url = f"https://www.youtube.com/watch?v={item['key']}"
            videos.append(Video(name=item["name"], type=item["type"], url=url))
        return videos

    @property
    def vote(self):
        """Return an instance of the `Vote` class. It is like a
        `NamedTuple` object with two attributes: `average` and
        `count`."""
        average = self._getdata("vote_average")
        count = self._getdata("vote_average")
        return Vote(average=average, count=count)

    @property
    def cast(self):
        """Return TV episode cast as list of tuples
        `(Person, Credit)`."""
        cast = []
        for item in self._getdata("credits")["cast"]:
            kwargs = {
                "credit_type": "cast",
                "department": "Acting",
                "job": "Actor",
            }
            credit = Credit(
                item["credit_id"],
                media_data=self.data,
                character=item["character"],
                **kwargs,
            )
            person = Person(item["id"], **item)
            cast.append((person, credit))
        return cast

    @property
    def crew(self):
        """Return TV episode crew as list of tuples
        `(Person, Credit)`."""
        crew = []
        for item in self._getdata("credits")["crew"]:
            kwargs = {
                "credit_type": "crew",
                "department": item["department"],
                "job": item["job"],
            }
            credit = Credit(item["credit_id"], media_data=self.data, **kwargs)
            person = Person(item["id"], **item)
            crew.append((person, credit))
        return crew

    @property
    def guest_stars(self):
        """Return TV episode guest stars as list of tuples
        `(Person, Credit)`."""
        guest_stars = []
        for item in self._getdata("credits")["guest_stars"]:
            credit = Credit(
                item["credit_id"],
                media_data=self.data,
                character=item.get("character"),
            )
            person = Person(item["id"], **item)
            guest_stars.append((person, credit))
        return guest_stars

    def get_all(self, **params):
        """Get all information about a TV episode in a single
        response."""
        methods = ",".join(URL.ALL_EPISODE_SECOND_SUFFIXES)
        all_data = self.get_details(
            **{"append_to_response": methods, **params}
        )
        self.data.update(all_data)
        return all_data

    def get_details(self, **params):
        """Get the primary TV episode details."""
        details = self._request(
            URL.EPISODE_DETAILS.format(
                **{
                    "show_id": self.show_id,
                    "season_number": self.season_number,
                    "episode_number": self.number,
                }
            ),
            **params,
        )
        self.data.update(details)
        return details

    def get_changes(self, **params):
        """Get the changes for a TV episode. By default only the
        last 24 hours are returned."""
        changes = self._request(
            URL.EPISODE_CHANGES.format(episode_id=self.tmdb_id), **params
        )
        self.data.update({"changes": changes})
        return changes

    def get_credits(self, **params):
        """Get the credits for TV episode."""
        movie_credits = self._request(
            URL.SEASON_CREDITS.format(
                **{
                    "show_id": self.show_id,
                    "season_number": self.season_number,
                    "episode_number": self.number,
                }
            ),
            **params,
        )
        self.data.update({"credits": movie_credits})
        return movie_credits

    def get_external_ids(self, **params):
        """Get the external ids for a TV episode."""
        external_ids = self._request(
            URL.EPISODE_EXTERNAL_IDS.format(
                **{
                    "show_id": self.show_id,
                    "season_number": self.season_number,
                    "episode_number": self.number,
                }
            ),
            **params,
        )
        self.data.update({"external_ids": external_ids})
        return external_ids

    def get_images(self, **params):
        """Get the images that belong to a TV episode."""
        images = self._request(
            URL.EPISODE_IMAGES.format(
                **{
                    "show_id": self.show_id,
                    "season_number": self.season_number,
                    "episode_number": self.number,
                }
            ),
            **params,
        )
        self.data.update({"images": images})
        return images

    def get_videos(self, **params):
        """Get the videos that have been added to a TV episode."""
        videos = self._request(
            URL.EPISODE_VIDEOS.format(
                **{
                    "show_id": self.show_id,
                    "season_number": self.season_number,
                    "episode_number": self.number,
                }
            ),
            **params,
        )
        self.data.update({"videos": videos})
        return videos

    def get_translations(self, **params):
        """Get a list of the translations that exist for a
        TV episode."""
        translations = self._request(
            URL.EPISODE_TRANSLATIONS.format(
                **{
                    "show_id": self.show_id,
                    "season_number": self.season_number,
                    "episode_number": self.number,
                }
            ),
            **params,
        )
        self.data.update({"translations": translations})
        return translations



import json
from urllib.parse import urlencode, urljoin
from urllib.request import Request, urlopen


def GET_total_pages_for(url, params):
    first_page = GET(url, page=1, **params)
    return first_page["total_pages"]


def GET(url, **params):
    params = urlencode({**params})
    with urlopen(f"{url}?{params}") as response:
        return json.loads(response.read().decode("utf-8"))


def GET_pages(url, params):
    for page in range(1, GET_total_pages_for(url, params) + 1):
        yield from GET(url, page=page, **params)["results"]


def POST(url, data, **params):
    headers = {"content-type": "application/json"}
    data = json.dumps(data).encode("utf-8")
    params = urlencode(params)
    request = Request(f"{url}?{params}", headers=headers, data=data)
    with urlopen(request) as response:
        return json.loads(response.read().decode("utf-8"))


def DELETE(url, data, **params):
    headers = {"content-type": "application/json"}
    data = json.dumps(data).encode("utf-8")
    url = f"{url}?{urlencode(params)}"
    request = Request(url, headers=headers, data=data, method="DELETE")
    with urlopen(request) as response:
        return json.loads(response.read().decode("utf-8"))



from operator import itemgetter
from typing import Iterator, List, Optional, Tuple
from importlib import import_module

import isle._urls as URL
from isle.objects._tmdb import TMDb


__all__ = ["Movie"]


def _import_person():
    global Person
    Person = import_module("isle.objects._person").Person


def _import_image():
    global Image
    Image = import_module("isle.objects._others").Image


def _import_language():
    global Language
    Language = import_module("isle.objects._others").Language


def _import_country():
    global Country
    Country = import_module("isle.objects._others").Country


def _import_company():
    global Company
    Company = import_module("isle.objects._company").Company


def _import_credit():
    global Credit
    Credit = import_module("isle.objects._others").Credit


def _import_vote():
    global Vote
    Vote = import_module("isle.objects._others").Vote


def _import_video():
    global Video
    Video = import_module("isle.objects._others").Video


def _import_genre():
    global Genre
    Genre = import_module("isle.objects._others").Genre


def _import_keyword():
    global Keyword
    Keyword = import_module("isle.objects._others").Keyword


def _import_all():
    _import_person()
    _import_image()
    _import_language()
    _import_country()
    _import_company()
    _import_credit()
    _import_vote()
    _import_video()
    _import_genre()
    _import_keyword()


class Movie(TMDb):
    """Represents a movie."""

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        _import_all()

    def _init(self):
        return self.get_all()

    @property
    def title(self):
        """Return titles of a movie in different languages.
        Each key is an ISO-3166-1 code (such as `"US"`, `"RU"`,
        etc.), and the value is the corresponding title.

        There is two special keys: `"original"` and `"default"`.
        The first one returns the original movie title and the
        second one depends on the location."""
        titles = {
            "default": self._getdata("title"),
            "original": self._getdata("original_title"),
        }
        for item in self._getdata("translations")["translations"]:
            code, title = item["iso_3166_1"], item["data"]["title"]
            if title:
                titles[code] = title
        return titles

    @property
    def overview(self):
        """Return the overviews of a movie in different
        languages. Each key is an ISO-3166-1 code (such as `"US"`,
        `"RU"`, etc.), and the value is the corresponding overview.

        There is a key `"default"`. It depends on the location."""
        overviews = {"default": self._getdata("overview")}
        for item in self._getdata("translations")["translations"]:
            code, overview = item["iso_3166_1"], item["data"]["overview"]
            if overview:
                overviews[code] = overview
        return overviews

    @property
    def tagline(self):
        """Return a movie slogan"""
        return self._getdata("tagline")

    @property
    def homepage(self):
        """Return a movie homepages. Each key is an ISO-3166-1
        code (such as `"US"`, `"RU"`, etc.) and the value is the
        corresponding homepage.

        There is a key `"default"`. It depends on the location.
        """
        default = self._getdata("homepage")
        pages = {"default": default} if default else {}
        for item in self._getdata("translations")["translations"]:
            code, page = item["iso_3166_1"], item["data"]["homepage"]
            if page:
                pages[code] = page
        return pages

    @property
    def year(self):
        """Return year of a movie."""
        release_date = self._getdata("release_date")
        if release_date:
            return int(release_date[0:4])
        else:
            return None

    @property
    def releases(self):
        """Return release dates of a movie in different
        countries. Each key is an ISO-3166-1 code (such as `"US"`,
        `"RU"`, etc.) and the value is the corresponding date in
        `YYYY-MM-DD` format."""

        def _remove_iso_639_1(d):
            if "iso_639_1" in d:
                del d["iso_639_1"]
            return d

        def _rename_to_date(d):
            if "release_date" in d:
                d["date"] = d["release_date"]
                del d["release_date"]
            return d

        def _default_note(d):
            if "note" not in d:
                d["note"] = ""
            return d

        dates = {}
        for item in self._getdata("release_dates")["results"]:
            release_dates = list(map(_remove_iso_639_1, item["release_dates"]))
            release_dates = list(map(_rename_to_date, release_dates))
            release_dates = list(map(_default_note, release_dates))
            dates[item["iso_3166_1"]] = release_dates
        return dates

    @property
    def is_adult(self):
        """Return `True` if a movie is for adults."""
        return self._getdata("adult")

    @property
    def backdrops(self):
        """Return backdrops that belong to a movie. Each item
        is an instance of the `Image` class."""
        backdrops = []
        for item in self._getdata("images")["backdrops"]:
            backdrops.append(Image(item, type_="backdrop"))
        return backdrops

    @property
    def posters(self):
        """Return posters that belong to a movie. Each item is
        an instance of the `Image` class."""
        posters = []
        for item in self._getdata("images")["posters"]:
            posters.append(Image(item, type_="poster"))
        return posters

    @property
    def languages(self):
        """Return the languages spoken in a movie. Each item is
        an instance of the `Language` class."""
        languages = []
        for code in map(
            itemgetter("iso_639_1"), self._getdata("spoken_languages")
        ):
            try:
                item = self._all_languages[code]
            except AttributeError:
                self._all_languages = self._get_all_languages()
                item = self._all_languages[code]
            languages.append(
                Language(
                    iso_639_1=code,
                    english_name=item["english_name"],
                    original_name=item["name"],
                )
            )
        return languages

    @property
    def countries(self):
        """Return production countries. Each item is an instance of
        the `Country` class."""
        countries = []
        for item in self._getdata("production_countries"):
            countries.append(Country(item["iso_3166_1"], item["name"]))
        return countries

    @property
    def popularity(self):
        """Return popularity of a movie on TMDb.

        See more:
        https://developers.themoviedb.org/3/getting-started/popularity."""
        return self._getdata("popularity")

    @property
    def revenue(self):
        """Return revenue of a movie."""
        return self._getdata("revenue")

    @property
    def budget(self):
        """Return budget of a movie."""
        return self._getdata("budget")

    @property
    def runtime(self):
        """Return a movie runtime."""
        return self._getdata("runtime")

    @property
    def status(self):
        """Return one of this values: `"Rumored"`, `"Planned"`,
        `"In Production"`, `"Post Production"`, `"Released"`,
        `"Canceled"`."""
        return self._getdata("status")

    @property
    def companies(self):
        """Return the production companies. Each item is an
        instance of the `Company` class."""
        companies = []
        for item in self._getdata("production_companies"):
            companies.append(Company(item["id"], **item))
        return companies

    @property
    def cast(self):
        """Return movie cast as list of tuples `(Person, Credit)`."""
        cast = []
        for item in self._getdata("credits")["cast"]:
            kwargs = {
                "credit_type": "cast",
                "media_type": "movie",
                "department": "Acting",
                "job": "Actor",
            }
            credit = Credit(
                item["credit_id"],
                media_data=self.data,
                character=item["character"],
                **kwargs,
            )
            person = Person(item["id"], **item)
            cast.append((person, credit))
        return cast

    @property
    def crew(self):
        """Return movie crew as list of tuples `(Person, Credit)`."""
        crew = []
        for item in self._getdata("credits")["crew"]:
            kwargs = {
                "credit_type": "crew",
                "media_type": "movie",
                "department": item["department"],
                "job": item["job"],
            }
            credit = Credit(item["credit_id"], media_data=self.data, **kwargs)
            person = Person(item["id"], **item)
            crew.append((person, credit))
        return crew

    @property
    def vote(self):
        """Return an instance of the `Vote` class. It is like a
        `NamedTuple` object with two attributes: `average` and
        `count`."""
        average = self._getdata("vote_average")
        count = self._getdata("vote_count")
        return Vote(average=average, count=count)

    @property
    def videos(self):
        """Return videos. Each item is an instance of the `Video`
        class."""
        videos = []
        for item in self._getdata("videos")["results"]:
            url = f"https://www.youtube.com/watch?v={item['key']}"
            videos.append(Video(name=item["name"], type=item["type"], url=url))
        return videos

    @property
    def genres(self):
        """Return a movie genres."""
        genres = []
        for item in self._getdata("genres"):
            genres.append(Genre(tmdb_id=item["id"], name=item["name"]))
        return genres

    @property
    def keywords(self):
        """Return a movie keywords."""
        keywords = []
        for item in self._getdata("keywords")["keywords"]:
            keywords.append(Keyword(item["id"], **item))
        return keywords

    @property
    def imdb_id(self):
        """Return the ID of a movie on IMDb."""
        return self._getdata("external_ids")["imdb_id"]

    @property
    def facebook_id(self):
        """Return the ID of a movie on Facebook."""
        return self._getdata("external_ids")["facebook_id"]

    @property
    def instagram_id(self):
        """Return the ID of a movie on Instagram."""
        return self._getdata("external_ids")["instagram_id"]

    @property
    def twitter_id(self):
        """Return the ID of a movie on Twitter."""
        return self._getdata("external_ids")["twitter_id"]

    def get_all(self, **params):
        """Get all information about a movie. This method
        makes only one API request."""
        methods = ",".join(URL.ALL_MOVIE_SECOND_SUFFIXES)
        all_data = self.get_details(
            **{"append_to_response": methods, **params}
        )
        self.data.update(all_data)
        return all_data

    def get_details(self, **params):
        """Get the primary information about a movie."""
        details = self._request(
            URL.MOVIE_DETAILS.format(movie_id=self.tmdb_id), **params
        )
        self.data.update(details)
        return details

    def get_alternative_titles(self, **params):
        """Get all of the alternative titles for a movie."""
        alternative_titles = self._request(
            URL.MOVIE_ALTERNATIVE_TITLES.format(movie_id=self.tmdb_id),
            **params,
        )
        self.data.update({"alternative_titles": alternative_titles})
        return alternative_titles

    def get_changes(self, **params):
        """Get the changes for a movie. By default only the
        last 24 hours are returned."""
        changes = self._request(
            URL.MOVIE_CHANGES.format(movie_id=self.tmdb_id), **params
        )
        self.data.update({"changes": changes})
        return changes

    def get_credits(self, **params):
        """Get the cast and crew for a movie."""
        credits = self._request(
            URL.MOVIE_CREDITS.format(movie_id=self.tmdb_id), **params
        )
        self.data.update({"credits": credits})
        return credits

    def get_external_ids(self, **params):
        """Get the external ids for a movie. Such as
        Facebook, Instagram, Twitter and IMDb"""
        external_ids = self._request(
            URL.MOVIE_EXTERNAL_IDS.format(movie_id=self.tmdb_id), **params
        )
        self.data.update({"external_ids": external_ids})
        return external_ids

    def get_images(self, **params):
        """Get the images that belong to a movie."""
        images = self._request(
            URL.MOVIE_IMAGES.format(movie_id=self.tmdb_id), **params
        )
        self.data.update({"images": images})
        return images

    def get_keywords(self, **params):
        """Get the keywords that have been added to a movie."""
        keywords = self._request(
            URL.MOVIE_KEYWORDS.format(movie_id=self.tmdb_id), **params
        )
        self.data.update({"keywords": keywords})
        return keywords

    def get_release_dates(self, **params):
        """Get the release date along with the certification
        for a movie."""
        release_dates = self._request(
            URL.MOVIE_RELEASE_DATES.format(movie_id=self.tmdb_id), **params
        )
        self.data.update({"release_dates": release_dates})
        return release_dates

    def get_videos(self, **params):
        """Get the videos that have been added to a movie."""
        videos = self._request(
            URL.MOVIE_VIDEOS.format(movie_id=self.tmdb_id), **params
        )
        self.data.update({"videos": videos})
        return videos

    def get_translations(self, **params):
        """Get a list of translations that have been created
        for a movie."""
        translations = self._request(
            URL.MOVIE_TRANSLATIONS.format(movie_id=self.tmdb_id), **params
        )
        self.data.update({"translations": translations})
        return translations

    def iter_recommendations(self, **params):
        """Get a list of recommended movies for a movie."""
        yield from self._iter_request(
            URL.MOVIE_RECOMMENDATIONS.format(movie_id=self.tmdb_id), **params
        )

    def iter_similar_movies(self, **params):
        """Get a list of recommended movies for a movie."""
        yield from self._iter_request(
            URL.MOVIE_SIMILAR.format(movie_id=self.tmdb_id), **params
        )

    def iter_reviews(self, **params):
        """Get the user reviews for a movie."""
        yield from self._iter_request(
            URL.MOVIE_REVIEWS.format(movie_id=self.tmdb_id), **params
        )

    def iter_lists(self, **params):
        """Get a list of lists that this movie belongs to."""
        yield from self._iter_request(
            URL.MOVIE_LISTS.format(movie_id=self.tmdb_id), **params
        )



import copy
from typing import NamedTuple
from importlib import import_module

import isle._urls as URL
from ._tmdb import TMDb
from .._config import tmdb_api_key


__all__ = ["Account", "TMDbList"]


def _import_movie():
    global Movie
    Movie = import_module("isle.objects._movie").Movie


def _import_show():
    global Show
    Show = import_module("isle.objects._show").Show


def _import_episode():
    global Episode
    Episode = import_module("isle.objects._show").Episode


def _import_language():
    global Language
    Language = import_module("isle.objects._others").Language


def _import_all():
    _import_movie()
    _import_show()
    _import_episode()
    _import_language()


class Account(TMDb):
    """Represents an user account."""

    class _Token(NamedTuple):
        id: str
        expires_at: str

    class _Session(NamedTuple):
        id: str
        is_guest: bool = False

    class SessionError(Exception):
        pass

    class TokenError(Exception):
        pass

    def __init__(self):
        _import_all()
        self.data = {}
        self._token = None
        self._session = None
        self.n_requests = 0

    def _init(self):
        self.get_details()

    def login(self, username, password):
        """Log in to the account."""
        self._create_token()
        self._validate_token(username, password)
        return self._create_session()

    def logout(self):
        """Logout from a session."""
        data = {"session_id": self._session_id}
        request = self._delete_request(URL.AUTH_DELETE_SESSION, data)
        self._token = self._session = None
        return request

    def login_as_guest(self):
        """Create a new guest session. Guest sessions are a type of
        session that will let a user rate movies and TV shows but
        not require them to have a TMDb user account. """
        request = self._request(URL.AUTH_GUEST_SESSION, api_key=tmdb_api_key())
        self._session = self._Session(
            id=request["guest_session_id"], is_guest=True
        )
        return request

    def _create_session(self):
        data = {"request_token": self._token_id}
        request = self._post_request(
            URL.AUTH_NEW_SESSION, data, api_key=tmdb_api_key()
        )
        self._session = self._Session(id=request["session_id"])
        return request

    def _create_token(self):
        """Create a temporary request token that can be used to
        validate a TMDb user login."""
        request = self._request(URL.AUTH_NEW_TOKEN, api_key=tmdb_api_key())
        self._token = self._Token(
            id=request["request_token"], expires_at=request["expires_at"]
        )
        return request

    def _validate_token(self, username, password):
        data = {
            "username": username,
            "password": password,
            "request_token": self._token_id,
        }
        request = self._post_request(
            URL.AUTH_VALIDATE_WITH_LOGIN, data, api_key=tmdb_api_key()
        )
        return request

    @property
    def _session_id(self):
        try:
            return self._session.id
        except AttributeError:
            raise self.SessionError("No session ID. Need to log in.")

    @property
    def _token_id(self):
        try:
            return self._token.id
        except AttributeError:
            raise self.TokenError("No request token ID. Need to log in.")

    @property
    def tmdb_id(self):
        return self._getdata("id")

    @property
    def default_language(self):
        """It is a ISO 639 1 code."""
        return self._getdata("iso_639_1")

    @property
    def fallback_language(self):
        """It is a ISO 3166 1 code."""
        return self._getdata("iso_3166_1")

    @property
    def name(self):
        return self._getdata("name")

    @property
    def username(self):
        return self._getdata("username")

    @property
    def include_adult(self):
        return self._getdata("include_adult")

    def get_details(self):
        """Get your account details."""
        request = self._request(
            URL.ACCOUNT_DETAILS,
            api_key=tmdb_api_key(),
            session_id=self._session_id,
        )
        self.data.update(request)
        return request

    def iter_lists(self, **params):
        """Get lists created by an account. Will include
        private lists if you are the owner."""
        params = {"session_id": self._session_id, **params}
        response = self._iter_request(
            URL.ACCOUNT_CREATED_LISTS.format(account_id=self.tmdb_id), **params
        )
        for item in response:
            yield TMDbList(item["id"], **item)

    def iter_favorite_movies(self, **params):
        """Get your favorite movies."""
        params = {"session_id": self._session_id, **params}
        response = self._iter_request(
            URL.ACCOUNT_FAVORITE_MOVIES.format(account_id=self.tmdb_id),
            **params,
        )
        for item in response:
            yield Movie(item["id"], **item)

    def iter_favorite_shows(self, **params):
        """Get your favorite TV shows."""
        params = {"session_id": self._session_id, **params}
        response = self._iter_request(
            URL.ACCOUNT_FAVORITE_SHOWS.format(account_id=self.tmdb_id),
            **params,
        )
        for item in response:
            yield Show(item["id"], **item)

    def iter_rated_movies(self, **params):
        """Get all the movies you have rated."""
        params = {"session_id": self._session_id, **params}
        response = self._iter_request(
            URL.ACCOUNT_RATED_MOVIES.format(account_id=self.tmdb_id), **params
        )
        for item in response:
            yield Movie(item["id"], **item)

    def iter_rated_shows(self, **params):
        """Get all the TV shows you have rated."""
        params = {"session_id": self._session_id, **params}
        response = self._iter_request(
            URL.ACCOUNT_RATED_SHOWS.format(account_id=self.tmdb_id), **params
        )
        for item in response:
            yield Show(item["id"], **item)

    def iter_rated_episodes(self, **params):
        """Get all the TV episodes you have rated."""
        params = {"session_id": self._session_id, **params}
        response = self._iter_request(
            URL.ACCOUNT_RATED_EPISODES.format(account_id=self.tmdb_id),
            **params,
        )
        for item in response:
            yield Episode(item["id"], **item)

    def iter_movie_watchlist(self, **params):
        """Get all the movies you have added to your watchlist."""
        params = {"session_id": self._session_id, **params}
        response = self._iter_request(
            URL.ACCOUNT_MOVIE_WATCHLIST.format(account_id=self.tmdb_id),
            **params,
        )
        for item in response:
            yield Movie(item["id"], **item)

    def iter_show_watchlist(self, **params):
        """Get all the TV shows you have added to your
        watchlist."""
        params = {"session_id": self._session_id, **params}
        response = self._iter_request(
            URL.ACCOUNT_SHOW_WATCHLIST.format(account_id=self.tmdb_id),
            **params,
        )
        for item in response:
            yield Show(item["id"], **item)

    def mark_as_favorite(self, item):
        """Mark a movie or TV show as a favorite item."""
        if not isinstance(item, (Movie, Show)):
            raise TypeError(
                f"An `item` must be `Movie` or `Show`, not a `{type(item)}`"
            )
        data = {
            "media_type": "movie" if isinstance(item, Movie) else "tv",
            "media_id": item.tmdb_id,
            "favorite": True,
        }
        return self._post_request(
            URL.ACCOUNT_MARK_AS_FAVORITE.format(account_id=self.tmdb_id),
            data,
            session_id=self._session_id,
        )

    def remove_from_favorites(self, item):
        """Remove a movie or TV show from your favorites."""
        if not isinstance(item, (Movie, Show)):
            raise TypeError(
                f"An `item` must be `Movie` or `Show`, not a `{type(item)}`"
            )
        data = {
            "media_type": "movie" if isinstance(item, Movie) else "tv",
            "media_id": item.tmdb_id,
            "favorite": False,
        }
        return self._post_request(
            URL.ACCOUNT_MARK_AS_FAVORITE.format(account_id=self.tmdb_id),
            data,
            session_id=self._session_id,
        )

    def add_to_watchlist(self, item):
        """Add a movie or TV show to your watchlist."""
        if not isinstance(item, (Movie, Show)):
            raise TypeError(
                f"An `item` must be `Movie` or `Show`, not a `{type(item)}`"
            )
        data = {
            "media_type": "movie" if isinstance(item, Movie) else "tv",
            "media_id": item.tmdb_id,
            "watchlist": True,
        }
        return self._post_request(
            URL.ACCOUNT_ADD_TO_WATCHLIST.format(account_id=self.tmdb_id),
            data,
            session_id=self._session_id,
        )

    def remove_from_watchlist(self, item):
        """Add a movie or TV show to your watchlist."""
        if not isinstance(item, (Movie, Show)):
            raise TypeError(
                f"An `item` must be `Movie` or `Show`, not a `{type(item)}`"
            )
        data = {
            "media_type": "movie" if isinstance(item, Movie) else "tv",
            "media_id": item.tmdb_id,
            "watchlist": False,
        }
        return self._post_request(
            URL.ACCOUNT_ADD_TO_WATCHLIST.format(account_id=self.tmdb_id),
            data,
            session_id=self._session_id,
        )

    def rate(self, item, value):
        """Rate a movie, TV show or TV episode. `value` must be
        between `0.5` and `10`"""
        if not isinstance(item, (Movie, Show)):
            raise TypeError(
                f"An `item` must be `Movie` or `Show`, not a `{type(item)}`"
            )
        if not 0.5 <= value <= 10:
            raise TypeError(f"A value must be between 0.5 and 10. Not {value}")
        data = {"value": value}
        if isinstance(item, Movie):
            return self._post_request(
                URL.RATE_MOVIE.format(movie_id=item.tmdb_id),
                data,
                session_id=self._session_id,
            )
        else:
            return self._post_request(
                URL.RATE_SHOW.format(show_id=item.tmdb_id),
                data,
                session_id=self._session_id,
            )

    def delete_rating(self, item):
        """Remove your rating for a movie, TV show or TV
        episode."""
        if not isinstance(item, (Movie, Show)):
            raise TypeError(
                f"An `item` must be `Movie` or `Show`, not a `{type(item)}`"
            )
        if isinstance(item, Movie):
            return self._delete_request(
                URL.DELETE_MOVIE_RATING.format(movie_id=item.tmdb_id),
                {},
                session_id=self._session_id,
            )
        else:
            return self._post_request(
                URL.DELETE_SHOW_RATING.format(show_id=item.tmdb_id),
                {},
                session_id=self._session_id,
            )

    # def create_list(self, name, desc, lang="en"):
    #     """Create a list."""
    #     data = {"name": name, "description": desc, "language": lang}
    #     r = self._post_request(URL.CREATE_LIST, data, session_id=self._session_id)
    #     return TMDbList(r["list_id"], name=name, description=desc, iso_639_1=lang)

    # def delete_list(self, list_): # TODO: raises HTTPError (it is a TMDb problem)
    #     """Delete a list."""
    #     return self._delete_request(
    #         URL.DELETE_LIST.format(list_id=list_.tmdb_id), {}, session_id=self._session_id
    #     )

    # def add_movie_to_list(self, list_, movie):
    #     """Add a movie to a list."""
    #     data = {"media_id": movie.tmdb_id}
    #     r = self._post_request(
    #         URL.ADD_MOVIE_TO_LIST.format(list_id=list_.tmdb_id),
    #         data,
    #         session_id=self._session_id,
    #         confirm=True,
    #     )
    #     if r["status_code"] == 12:
    #         list_._changed = True
    #     return r

    # def remove_movie_from_list(self, list_, movie):
    #     """Remove a movie from a list."""
    #     data = {"media_id": movie.tmdb_id}
    #     r = self._post_request(
    #         URL.REMOVE_MOVIE_FROM_LIST.format(list_id=list_.tmdb_id),
    #         data,
    #         session_id=self._session_id,
    #     )
    #     if r["status_code"] == 13:
    #         list_._changed = True
    #     return r

    # def clear_list(self, list_):
    #     """Clear all of the items from a list."""
    #     r = self._post_request(
    #         URL.CLEAR_LIST.format(list_id=list_.tmdb_id),
    #         None,
    #         session_id=self._session_id,
    #     )
    #     if r["status_code"] == 12:
    #         list_._changed = True
    #     return r


class TMDbList(TMDb):
    """Represents a list."""

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        _import_all()
        self._changed = False

    def _init(self):
        self.get_details()

    def _getdata(self, key):
        if key not in self.data or self._changed:
            self._init()
            self._changed = False
        return copy.deepcopy(self.data[key])

    @property
    def name(self):
        """The name of a list."""
        return self._getdata("name")

    @property
    def description(self):
        """The description of a list."""
        return self._getdata("description")

    @property
    def creator(self):
        return self._getdata("created_by")

    @property
    def n_favorites(self):
        return self._getdata("favorite_count")

    @property
    def items(self):
        acc = []
        for item in self._getdata("items"):
            if item["media_type"] == "movie":
                Obj = Movie
            elif item["media_type"] == "tv":
                Obj = Show
            acc.append(Obj(item["id"], **item))
        return acc

    @property
    def language(self):
        code = self._getdata("iso_639_1")
        try:
            item = self._all_languages[code]
        except AttributeError:
            self._all_languages = self._get_all_languages()
            item = self._all_languages[code]
        return Language(
            iso_639_1=code,
            english_name=item["english_name"],
            original_name=item["name"],
        )

    def get_details(self, **params):
        details = self._request(
            URL.LIST_DETAILS.format(list_id=self.tmdb_id), **params
        )
        self.n_requests += 1
        self.data.update(details)
        return details

    def has_movie(self, movie):
        r = self._request(
            URL.LIST_CHECK_MOVIE_STATUS.format(list_id=self.tmdb_id),
            movie_id=movie.tmdb_id,
        )
        return r["item_present"]



from ._account import *
from ._company import *
from ._movie import *
from ._others import *
from ._person import *
from ._show import *

__all__ = [
    "Account",
    "Movie",
    "Show",
    "Person",
    "Company",
    "Season",
    "Episode",
]



import copy
from abc import ABC, abstractmethod

import isle._urls as URL

from .._config import tmdb_api_key
from .._requests import DELETE, GET, POST, GET_pages


class TMDb(ABC):
    def __init__(self, tmdb_id: int, **kwargs):
        self.data = {"id": tmdb_id, **kwargs}
        self.tmdb_id = self.data["id"]
        self.n_requests = 0

    @abstractmethod
    def _init(self):
        pass

    def _getdata(self, key):
        if key not in self.data:
            self._init()
        return copy.deepcopy(self.data[key])

    def _request(self, url: str, **params):
        self.n_requests += 1
        return GET(url, **{"api_key": tmdb_api_key(), **params})

    def _iter_request(self, url: str, **params):
        self.n_requests += 1
        return GET_pages(url, {"api_key": tmdb_api_key(), **params})

    def _post_request(self, url, data, **params):
        params = {"api_key": tmdb_api_key(), **params}
        request = POST(url, data, **params)
        self.n_requests += 1
        return request

    def _delete_request(self, url, data, **params):
        params = {"api_key": tmdb_api_key(), **params}
        request = DELETE(url, data, **params)
        self.n_requests += 1
        return request

    def _get_all_languages(self):
        data = self._request(
            URL.LANGUAGES_CONFIGURATION, **{"api_key": tmdb_api_key()}
        )
        languages = {}
        for item in data:
            iso_639_1 = item["iso_639_1"]
            del item["iso_639_1"]
            languages[iso_639_1] = item
        return languages

    def __repr__(self):
        return f"{type(self).__name__}({self.tmdb_id})"
